#include <stdio.h>
#include <unistd.h>

#include "nvsoftboot.h"
#include "bios.h"
#include "initcmds.h"
#include "parser.h"
#include "nv.h"

bool_t init_io_restrict_prog(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    
    /* INIT_IO_RESTRICT_PROG   opcode: 0x32
     * 
     * offset      (8  bit): opcode
     * offset + 1  (16 bit): CRTC reg
     * offset + 3  (8  bit): CRTC index
     * offset + 4  (8  bit): and mask
     * offset + 5  (8  bit): shift right
     * offset + 6  (8  bit): number of configurations
     * offset + 7  (32 bit): register
     * offset + 11 (32 bit): configuration 1
     * ...
     * 
     * Starting at offset + 11 there are "number of configurations"
     * 32 bit values. To find out which configuration value to use
     * read "CRTC reg" on the CRTC controller with index "CRTC index"
     * and bitwise AND this value with "and mask" and then bit shift the
     * result "shift right" bits to the right.
     * Assign "register" with appropriate configuration value.
     */
    
    uint16_t crtcreg = *((uint16_t *) (&bios->data[offset + 1]));
    uint8_t  index = *((uint8_t *) (&bios->data[offset + 3]));
    uint8_t and = *((uint8_t *) (&bios->data[offset + 4]));
    uint8_t shiftr = *((uint8_t *) (&bios->data[offset + 5]));
    uint8_t nr = *((uint8_t *) (&bios->data[offset + 6]));
    uint32_t reg = *((uint32_t *) (&bios->data[offset + 7]));
    uint8_t configuration;
    uint32_t configval, tmp;
   	
	if (iexec->execute) {
		DEBUGPRINT(6, "0x%04X: CRTC REG: 0x%04X, INDEX: 0x%02X\n", offset, 
				crtcreg, index, reg);

		nv_wr_pcio(crtcreg, index);
		configuration = (nv_rd_pcio(crtcreg + 1) & and) >> shiftr;

		DEBUGPRINT(6, "0x%04X: CONFIGURATION TO USE: 0x%02X\n", 
				offset, configuration);

		if (configuration <= nr) {


			configval = 
				*((uint32_t *) (&bios->data[offset + 11 + configuration * 4]));

			DEBUGPRINT(5, "0x%04X: REG: 0x%08X, VALUE: 0x%08X\n", offset, 
					reg, configval);
			
			if (nv32_rd(reg, &tmp)) {
				DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%08X\n", 
						offset, tmp);
			}
#ifdef PERFORM_WRITE 
			nv32_wr(reg, configval);
#endif
		}
	}
    return TRUE;
}

bool_t init_repeat(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    
    
    uint8_t repeats = *((uint8_t *) (&bios->data[offset + 1]));
    uint8_t i;
    
    if (iexec->execute) {
		DEBUGPRINT(5, "0x%04X: REPEATING FOLLOWING SEGMENT %d TIMES.\n", 
				offset, repeats);

		iexec->repeat = TRUE;

		for (i = 0; i < repeats - 1; i++)
			parse_init_table(bios, offset + 2, iexec);

		iexec->repeat = FALSE;
	}
    return TRUE;
}

bool_t init_end_repeat(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    if (iexec->repeat)
        return FALSE;
    
    return TRUE;
}

bool_t init_copy(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    /* XXX: double check this... */
    uint32_t reg = *((uint32_t *) (&bios->data[offset + 1]));
    uint8_t shift = *((uint8_t *) (&bios->data[offset + 5]));
    uint8_t and1 = *((uint8_t *) (&bios->data[offset + 6]));
    uint16_t crtcreg = *((uint16_t *) (&bios->data[offset + 7]));
    uint8_t index = *((uint8_t *) (&bios->data[offset + 9]));
    uint8_t and2 = *((uint8_t *) (&bios->data[offset + 10]));
    uint32_t data;
    uint8_t crtcdata;
   	
	if (iexec->execute) {
		if (nv32_rd(reg, &data)) {
			if (shift < 0x80) 
				data >>= shift;
			else
				data <<= (0x100 - shift);

			data &= and1;
			nv_wr_pcio(crtcreg, index);
			crtcdata = (nv_rd_pcio(crtcreg + 1) & and2) | (uint8_t) data;

			DEBUGPRINT(5, 
					"0x%04X: CRTC REG: 0x%04X, INDEX: 0x%04X, VALUE: 0x%02X\n"
					, offset, crtcreg, index, crtcdata);

			DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%02X\n", offset, 
					nv_rd_pcio(crtcreg + 1));
#ifdef PERFORM_WRITE 
			printf("WRITE IS PERFORMED\n");
			nv_wr_pcio(crtcreg + 1, crtcdata);
#endif
		}
	}
    return TRUE;
}

bool_t init_not(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    if (iexec->execute) { 
        DEBUGPRINT(3, "0x%04X: ------ SKIPPING FOLLOWING COMMANDS  ------\n",
                offset);
    } else {
        DEBUGPRINT(3, "0x%04X: ------ EXECUTING FOLLOWING COMMANDS ------\n",
                offset);
    }

    iexec->execute = !iexec->execute;
    return TRUE;
}

bool_t init_io_flag_condition(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    uint8_t cond = *((uint8_t *) (&bios->data[offset + 1]));
    uint16_t crtcreg = *((uint16_t *) 
            (&bios->data[bios->io_flag_condition_offset + 
             cond * IO_FLAG_CONDITION_SIZE]));
    uint8_t index = *((uint8_t *) 
            (&bios->data[bios->io_flag_condition_offset + 
             cond * IO_FLAG_CONDITION_SIZE + 2]));
    uint8_t and1 = *((uint8_t *) 
            (&bios->data[bios->io_flag_condition_offset + 
             cond * IO_FLAG_CONDITION_SIZE + 3]));
    uint8_t shift = *((uint8_t *) 
            (&bios->data[bios->io_flag_condition_offset + 
             cond * IO_FLAG_CONDITION_SIZE + 4]));
    uint16_t offs = *((uint16_t *) 
            (&bios->data[bios->io_flag_condition_offset + 
             cond * IO_FLAG_CONDITION_SIZE + 5]));
    uint8_t and2 = *((uint8_t *) 
            (&bios->data[bios->io_flag_condition_offset + 
             cond * IO_FLAG_CONDITION_SIZE + 7]));
    uint8_t cmpval = *((uint8_t *) 
            (&bios->data[bios->io_flag_condition_offset + 
             cond * IO_FLAG_CONDITION_SIZE + 8]));
    
    uint8_t data;
	
	if (iexec->execute) {
	
		nv_wr_pcio(crtcreg, index);
		data = nv_rd_pcio(crtcreg + 1);
		data &= and1;
		offs += (data >> shift);
		data = *((uint8_t *) (&bios->data[offs]));
		data &= and2;

		DEBUGPRINT(5, 
				"0x%04X: CHECKING IF DATA: %02X equals COND: %02X\n", offset, 
				data, cmpval);

		if (data == cmpval) {
			DEBUGPRINT(5, 
					"0x%04X: CONDITION FULFILLED - CONTINUING TO EXECUTE\n", 
					offset);
		} else {
			DEBUGPRINT(5, "0x%04X: CONDITION IS NOT FULFILLED.\n", offset);
			DEBUGPRINT(3, 
					"0x%04X: ------ SKIPPING FOLLOWING COMMANDS  ------\n");
			iexec->execute = FALSE;     
		}
	}

    return TRUE;
}

bool_t init_io_restrict_pll(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{

    DEBUGPRINT(2, "0x%04X: [ NOT YET IMPLEMENTED ]\n", offset);
    /* XXX: this needs to be confirmed... NOT CORRECT */
    //init_io_restrict_prog(bios, offset, iexec);

    
    uint16_t crtcreg = *((uint16_t *) (&bios->data[offset + 1]));
    uint8_t  index = *((uint8_t *) (&bios->data[offset + 3]));
    uint8_t and = *((uint8_t *) (&bios->data[offset + 4]));
    uint8_t shiftr = *((uint8_t *) (&bios->data[offset + 5]));
    uint8_t nr = *((uint8_t *) (&bios->data[offset + 6]));
    uint32_t reg = *((uint32_t *) (&bios->data[offset + 7]));
    uint8_t configuration;
    uint32_t configval, tmp;
/*    
    DEBUGPRINT(6, "0x%04X: CRTC REG: 0x%04X, INDEX: 0x%02X\n", offset, 
            crtcreg, index, reg);
    
    nv_wr_pcio(crtcreg, index);
    configuration = (nv_rd_pcio(crtcreg + 1) & and) >> shiftr;
    
    DEBUGPRINT(6, "0x%04X: CONFIGURATION TO USE: 0x%02X\n", 
            offset, configuration);
    
    if (configuration <= nr) {
        
        if (DEBUGLEVEL >= 6 && nv32_rd(reg, &configval)) 
            DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%08X\n", 
                    offset, configval);
        
        configval = 
            *((uint32_t *) (&bios->data[offset + 11 + configuration * 4]));

        DEBUGPRINT(5, "0x%04X: REG: 0x%08X, VALUE: 0x%08X\n", offset, 
                reg, configval);
    }
*/
	if (iexec->execute) {    
		switch (reg) {
			case 0x00004004:
				configval = 0x01014E07;
				break;
			case 0x00004024:
				configval = 0x13030E02;
				break;
		}

		DEBUGPRINT(5, "0x%04X: REG: 0x%08X, VALUE: 0x%08X\n", offset, 
				reg, configval);

		if (DEBUGLEVEL >= 6 && nv32_rd(reg, &tmp))
			DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%08X\n", 
					offset, tmp);

#ifdef PERFORM_WRITE 
		nv32_wr(reg, configval);
#endif

	}
    return TRUE;
}

bool_t init_pll(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
 
    uint32_t reg = *((uint32_t *) (&bios->data[offset + 1]));
    uint32_t val = *((uint32_t *) (&bios->data[offset + 5]));
    uint32_t configval, tmp;

	if (iexec->execute) {
		switch (reg) {
			case 0x00680508:
				configval = 0x00011F05;
				break;
		}

		DEBUGPRINT(5, "0x%04X: REG: 0x%08X, VALUE: 0x%08X\n", offset, reg,
				configval);


		if (DEBUGLEVEL >= 6 && nv32_rd(reg, &tmp)) 
			DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%08X\n", 
					offset, tmp);

#ifdef PERFORM_WRITE 
		nv32_wr(reg, configval);
#endif

		//DEBUGPRINT(2, "0x%04X: [ NOT YET IMPLEMENTED ]\n", offset);
	}
    return TRUE;
}

bool_t init_cr_idx_adr_latch(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    
    uint8_t crtcindex = *((uint8_t *) (&bios->data[offset + 1]));
    uint8_t crtcdata = *((uint8_t *) (&bios->data[offset + 2]));
    uint8_t initial_index = *((uint8_t *) (&bios->data[offset + 3]));
	uint8_t entries = *((uint8_t *) (&bios->data[offset + 4]));
    uint8_t data;
	int i;
	
	if (iexec->execute) {
		for (i = 0; i < entries; i++) {
			nv_set_crtc_index(crtcindex);

			DEBUGPRINT(5, "0x%04X: CRTC INDEX: %02X    DATA: %02X\n", offset,
					crtcindex, initial_index + i);

			DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%02X\n", offset, 
					nv_rd_crtc_data());

#ifdef PERFORM_WRITE 
			nv_wr_crtc_data(initial_index + i);
#endif
			
			nv_set_crtc_index(crtcdata);

			data = *((uint8_t *) (&bios->data[offset + 5 + i]));

			DEBUGPRINT(5, "0x%04X: CRTC INDEX: %02X    DATA: %02X\n", offset,
					crtcdata, data);

			DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%02X\n", offset, 
					nv_rd_crtc_data());

#ifdef PERFORM_WRITE 
			nv_wr_crtc_data(data);
#endif

		}
	}
	return TRUE;
}


bool_t init_cr(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    
    
    /* XXX: IS THIS CORRECT? check the typecast .. probably wrong */

    uint8_t index = *((uint32_t *) (&bios->data[offset + 1]));
    uint8_t and = *((uint8_t *) (&bios->data[offset + 2]));
    uint8_t or = *((uint8_t *) (&bios->data[offset + 3]));
    uint8_t data;
	
	if (iexec->execute) {
		nv_set_crtc_index(index);
		data = (nv_rd_crtc_data() & and) | or;
		//printf("and: 0x%02x    or: 0x%02x\n", and, or);
		DEBUGPRINT(5, "0x%04X: CRTC INDEX: 0x%02X, VALUE: 0x%02X\n", offset, 
				index, data);

		DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%02X\n", offset, 
				nv_rd_crtc_data());

#ifdef PERFORM_WRITE 
		nv_wr_crtc_data(data);
#endif

	}
    return TRUE;
    
}

bool_t init_zm_cr(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    
    /* INIT_ZM_CR   opcode: 0x53
     * 
     * offset      (8  bit): opcode
     * offset + 1  (8  bit): CRTC index
     * offset + 2  (8  bit): value
     * 
     * Assign "value" to CRTC register with index "CRTC index".
     */
    
    uint8_t index = *((uint32_t *) (&bios->data[offset + 1]));
    uint8_t value = *((uint8_t *) (&bios->data[offset + 2]));
	
	if (iexec->execute) {
		DEBUGPRINT(5, "0x%04X: CRTC INDEX: 0x%02X, VALUE: 0x%02X\n", offset, 
				index, value);

		nv_set_crtc_index(index);

		DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%02X\n", offset, 
				nv_rd_crtc_data());

#ifdef PERFORM_WRITE 
		nv_wr_crtc_data(value);
#endif
	}
    return TRUE;
}

bool_t init_zm_cr_group(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{

    /* INIT_ZM_CR   opcode: 0x54
     * 
     * offset      (8  bit): opcode
     * offset + 1  (8  bit): number of groups (index, value)
     * offset + 2  (8  bit): index 1
     * offset + 3  (8  bit): value 1
     * ...
     * 
     * Assign "value n" to CRTC register with index "index n".
     */
    
    uint8_t nr = *((uint8_t *) (&bios->data[offset + 1]));
    uint8_t index, value;
    int i;
    
	if (iexec->execute) {
		for (i = 0; i < nr; i++) {
			index = *((uint8_t *) (&bios->data[offset + 2 + 2 * i]));
			value = *((uint8_t *) (&bios->data[offset + 2 + 2 * i + 1]));

			DEBUGPRINT(5, "0x%04X: CRTC INDEX: 0x%02X, VALUE: 0x%02X\n", offset,
					index, value);

			nv_set_crtc_index(index);

			DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%02X\n", offset,
					nv_rd_crtc_data());
#ifdef PERFORM_WRITE
			nv_wr_crtc_data(value);
#endif
		}
	}
    return TRUE;
}

bool_t init_condition_time(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    /* My BIOS does not use this command. */
    DEBUGPRINT(2, "0x%04X: [ NOT YET IMPLEMENTED ]\n", offset);

    return FALSE;
}

bool_t init_zm_reg_sequence(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    
    /* INIT_ZM_REG_SEQUENCE   opcode: 0x58
     * 
     * offset      (8  bit): opcode
     * offset + 1  (32 bit): register base
     * offset + 5  (8  bit): nr
     * offset + 6  (32 bit): value to assign "register base" + 4
     * ...
     * 
     * Initialzies a sequence of "nr" registers starting at "register base".
     */
    
    uint32_t reg = *((uint32_t *) (&bios->data[offset + 1]));
    uint32_t nr = *((uint8_t *) (&bios->data[offset + 5]));
    uint32_t data;
    uint32_t tmp;
	int i;
    
   	if (iexec->execute) { 
		for (i = 0; i < nr; i++) {
			data = *((uint32_t *) (&bios->data[offset + 6 + i * 4]));
			DEBUGPRINT(5, "0x%04X: REG: 0x%08X, VALUE: 0x%08X\n", offset,
					reg + i * 4, data);
			
			if (nv32_rd(reg + i * 4, &tmp)) { 
				DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%08X\n", 
						offset, tmp);
			}

#ifdef PERFORM_WRITE
			nv32_wr(reg + i * 4, data);
#endif

		}
	}
    return TRUE;
}

bool_t init_sub_direct(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    
    /* INIT_SUB_DIRECT   opcode: 0x5B
     * 
     * offset      (8  bit): opcode
     * offset + 1  (16 bit): subroutine offset (in bios)
     *
     * Calls a subroutine that will execute commands until INIT_DONE
     * is found. 
     */
    
    uint16_t sub_offset = *((uint16_t *) (&bios->data[offset + 1]));
       
	if (iexec->execute) {
		DEBUGPRINT(5, "0x%04X: EXECUTING SUB-ROUTINE AT: 0x%04X\n", 
				offset, sub_offset);

		parse_init_table(bios, sub_offset, iexec);

		DEBUGPRINT(5, "0x%04X: END OF SUB-ROUTINE\n", offset);
	}
    return TRUE;
}

bool_t init_copy_nv_reg(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{   
    
	uint32_t srcreg = *((uint32_t *) (&bios->data[offset + 1]));
    int8_t shift = *((uint8_t *) (&bios->data[offset + 5]));
    uint32_t and1 = *((uint32_t *) (&bios->data[offset + 6]));
    uint32_t xor = *((uint32_t *) (&bios->data[offset + 10]));
    uint32_t dstreg = *((uint32_t *) (&bios->data[offset + 14]));
    uint32_t and2 = *((uint32_t *) (&bios->data[offset + 18]));
    uint32_t srcdata;
	uint32_t dstdata;
	
	if (iexec->execute) {
		nv32_rd(srcreg, &srcdata);

		if (shift > 0)
			srcdata >>= shift;
		else
			srcdata <<= shift;

		srcdata = (srcdata & and1) ^ xor;

		nv32_rd(dstreg, &dstdata);
		dstdata &= and2;

		dstdata |= srcdata;

		uint32_t tmp;		
		nv32_rd(dstreg, &tmp);

		DEBUGPRINT(5, "0x%04X: REG: 0x%08X, VALUE: 0x%08X\n", offset, dstreg, 
				dstdata);

		DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%08X\n", offset, tmp);

#ifdef PERFORM_WRITE
		nv32_wr(dstreg, dstdata);
#endif

	}
	return TRUE;
}

bool_t init_zm_index_io(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    uint16_t crtcreg = *((uint16_t *) (&bios->data[offset + 1]));
    uint8_t index = *((uint8_t *) (&bios->data[offset + 3]));
    uint8_t value = *((uint8_t *) (&bios->data[offset + 4]));
    
  	
	if (iexec->execute) {
		DEBUGPRINT(5, 
				"0x%04X: CRTC REG: 0x%04X, INDEX: 0x%04X, VALUE: 0x%02X\n", 
				offset, crtcreg, index, value);

		nv_wr_pcio(crtcreg, index);

		DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%02X\n", offset, 
				nv_rd_pcio(crtcreg + 1));
	
#ifdef PERFORM_WRITE
		nv_wr_pcio(crtcreg + 1, value);
#endif
	}
    return TRUE;
}

bool_t init_compute_mem(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{

	uint16_t ramcfg = *((uint16_t *) (&bios->data[bios->ram_table_offset]));
    uint32_t pfb_debug;
    uint32_t strapinfo;
    uint32_t ramcfg2;
    
	if (iexec->execute) {
		nv32_rd(0x00101000, &strapinfo);
		nv32_rd(0x00100080, &pfb_debug);

		DEBUGPRINT(5, "STRAPINFO: 0x%08X\n", strapinfo);
		DEBUGPRINT(5, "PFB_DEBUG: 0x%08X\n", pfb_debug);
		DEBUGPRINT(5, "RAM CFG: 0x%04X\n", ramcfg);

		pfb_debug &= 0xffffffef;
		strapinfo >>= 2;
		strapinfo &= 0x0000000f;
		ramcfg2 = *((uint16_t *) 
				(&bios->data[bios->ram_table_offset + (2 * strapinfo)])); 

		DEBUGPRINT(5, "AFTER MANIPULATION\n");
		DEBUGPRINT(5, "STRAPINFO: 0x%08X\n", strapinfo);
		DEBUGPRINT(5, "PFB_DEBUG: 0x%08X\n", pfb_debug);
		DEBUGPRINT(5, "RAM CFG2: 0x%08X\n", ramcfg2);

		DEBUGPRINT(2, "0x%04X: [ NOT YET IMPLEMENTED ]\n", offset);


		uint32_t reg1;
		uint32_t reg2;

		nv32_rd(0x00100200, &reg1);
		nv32_rd(0x0010020C, &reg2);

		DEBUGPRINT(5, "0x00100200: 0x%08X\n", reg1);
		DEBUGPRINT(5, "0x0010020C: 0x%08X\n", reg2);


	}
    return TRUE;
}

bool_t init_reset(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    
    uint32_t reg = *((uint32_t *) (&bios->data[offset + 1]));
    uint32_t value1 = *((uint32_t *) (&bios->data[offset + 5]));
    uint32_t value2 = *((uint32_t *) (&bios->data[offset + 9]));
    
	if (iexec->execute) {
		DEBUGPRINT(5, "0x%04X: REG: 0x%08X, VALUE: 0x%08X\n", 
				offset, reg, value1);
		DEBUGPRINT(5, "0x%04X: REG: 0x%08X, VALUE: 0x%08X\n", 
				offset, reg, value2);

		if (DEBUGLEVEL >= 6) {
			uint32_t tmpval;
			nv32_rd(reg, &tmpval);
			DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%08X\n", offset, tmpval);
			/*
			nv32_rd(PCICFG(PCICFG_ROMSHADOW), &tmpval);
			DEBUGPRINT(5, "0x%04X: PCICFG_ROMSHADOW: 0x%02X\n", offset, tmpval);
			*/
		}
#ifdef PERFORM_WRITE
		nv32_wr(reg, value1);
		nv32_wr(reg, value2);
#endif

	}
    /* PCI Config space init needs to be added here. */
    /*
    if (nv32_rd(PCICFG(PCICFG_ROMSHADOW), value1))
        nv32_wr(PCICFG(PCICFG_ROMSHADOW), value1 & 0xfffffffe)   
    
    */
    return TRUE;
}

bool_t init_sub(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    uint8_t sub = *((uint8_t *) (&bios->data[offset + 1]));
    
    if (iexec->execute) {
		DEBUGPRINT(5, "0x%04X: EXECUTING SUB-SCRIPT: %d\n", offset, sub);

		parse_init_table(bios, 
				*((uint16_t *) (&bios->data[bios->init_tbls_offset + sub * 2])),
				iexec);

		DEBUGPRINT(5, "0x%04X: END OF SUB-SCRIPT\n", offset, sub);
	}
    return TRUE;
}

bool_t init_nv_reg(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    
    /* INIT_NV_REG   opcode: 0x6E
     * 
     * offset      (8  bit): opcode
     * offset + 1  (32 bit): register
     * offset + 5  (32 bit): and mask
     * offset + 9  (32 bit): or with
     *
     * Assign "register" to (REGVAL(register) & "and mask") | "or with";
     */
    
    uint32_t reg = *((uint32_t *) (&bios->data[offset + 1]));
    uint32_t and = *((uint32_t *) (&bios->data[offset + 5]));
    uint32_t or = *((uint32_t *) (&bios->data[offset + 9]));
    uint32_t data;
    status_t status;
	
	if (iexec->execute) {
    
		/* end temp test */
		if ((status = nv32_rd(reg, &data))) {
			data = (data & and) | or;
			DEBUGPRINT(5, "0x%04X: REG: 0x%08X, VALUE: 0x%08X\n", offset, 
					reg, data);

			if (DEBUGLEVEL >= 6 && status) {
				uint32_t tmpval;
				nv32_rd(reg, &tmpval);
				DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%08X\n",
						offset, tmpval);
			}

#ifdef PERFORM_WRITE
			nv32_wr(reg, data);
#endif
			/* Assign: reg = data */
		}
	}
    return TRUE;
}

bool_t init_macro(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    /* XXX: Not sure this is correct... */
    
    uint8_t macro = *((uint8_t *) (&bios->data[offset + 1]));
    uint32_t reg = 
        *((uint32_t *) (&bios->data[bios->macro_offset + macro * MACRO_SIZE]));
    uint32_t value =
        *((uint32_t *) (&bios->data[bios->macro_offset + macro * MACRO_SIZE + 4]));
    
    if (iexec->execute) {
		DEBUGPRINT(5, "0x%04X: EXECUTING MACRO: 0x%02X\n", offset, macro);
		DEBUGPRINT(5, "0x%04X: REG: 0x%08X, VALUE: 0x%08X\n", offset, reg,
				value);

		if (DEBUGLEVEL >= 6) {
			uint32_t tmpval;
			nv32_rd(reg, &tmpval);
			DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%08X\n",
					offset, tmpval);
		}

#ifdef PERFORM_WRITE
		nv32_wr(reg, value);
#endif
		
	}
    return TRUE;

}

bool_t init_done(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{

}

bool_t init_resume(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    if (!iexec->execute) {
        iexec->execute = TRUE;;
        DEBUGPRINT(3, "0x%04X: ---- EXECUTING FOLLOWING COMMANDS ----\n",
                offset);
    }
    return TRUE;
}

bool_t init_time(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    
    /* INIT_TIME   opcode: 0x74
     * 
     * offset      (8  bit): opcode
     * offset + 1  (16 bit): time
     * 
     * Sleep for "time" microseconds.
     */
    
    uint16_t time = *((uint16_t *) (&bios->data[offset + 1]));
	
	if (iexec->execute) {
    	DEBUGPRINT(5, "0x%04X: Sleeping for 0x%04X microseconds.\n", 
				offset, time);
    
    	usleep(time);
	}
    return TRUE;
}

bool_t init_condition(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    uint8_t cond = *((uint8_t *) (&bios->data[offset + 1]));
    uint32_t reg = 
        *((uint32_t *) 
                (&bios->data[bios->condition_offset + cond * CONDITION_SIZE]));
    uint32_t and = 
        *((uint32_t *) 
                (&bios->data[bios->condition_offset + cond * CONDITION_SIZE + 4]));
    uint32_t cmpval = 
        *((uint32_t *) 
                (&bios->data[bios->condition_offset + cond * CONDITION_SIZE + 8]));
    uint32_t data;
    
	if (iexec->execute) {
		if (nv32_rd(reg, &data)) {
			data &= and;

			DEBUGPRINT(5, 
					"0x%04X: CHECKING IF REGVAL: 0x%08X equals COND: 0x%08X\n",
					offset, data, cmpval);

			if (data == cmpval) {
				DEBUGPRINT(5, 
						"0x%04X: CONDITION FULFILLED - CONTINUING TO EXECUTE\n",
						offset);
			} else {
				DEBUGPRINT(5, "0x%04X: CONDITION IS NOT FULFILLED.\n", offset);
				DEBUGPRINT(3, 
						"0x%04X: ------ SKIPPING FOLLOWING COMMANDS  ------\n");
				iexec->execute = FALSE;     
			}

		}
	}
    return TRUE;
}

bool_t init_index_io(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    
    /* INIT_INDEX_IO   opcode: 0x78
     * 
     * offset      (8  bit): opcode
     * offset + 1  (16 bit): CRTC reg
     * offset + 3  (8  bit): CRTC index
     * offset + 4  (8  bit): and mask
     * offset + 5  (8  bit): or with
     * 
     * 
     */
    
    uint16_t crtcreg = *((uint16_t *) (&bios->data[offset + 1]));
    uint8_t index = *((uint8_t *) (&bios->data[offset + 3]));
    uint8_t and  = *((uint8_t *) (&bios->data[offset + 4]));
    uint8_t or = *((uint8_t *) (&bios->data[offset + 5]));
    uint8_t data;
  	

	if (iexec->execute) {
		nv_wr_pcio(crtcreg, index);
		/* data at reg + 1 */
		data = (nv_rd_pcio(crtcreg + 1) & and) | or;

		DEBUGPRINT(5, 
				"0x%04X: CRTC REG: 0x%04X, INDEX: 0x%04X, VALUE: 0x%02X\n", 
				offset, crtcreg, index, data);
		DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%02X\n", offset, 
				nv_rd_pcio(crtcreg + 1));

#ifdef PERFORM_WRITE
		nv_wr_pcio(crtcreg + 1, data);
#endif
		
	}
	return TRUE;
	
}

bool_t init_zm_reg(bios_t *bios, uint16_t offset, init_exec_t *iexec)
{
    
    /* INIT_ZM_REG   opcode: 0x7A
     * 
     * offset      (8  bit): opcode
     * offset + 1  (32 bit): register
     * offset + 5  (32 bit): value
     * 
     * Assign "register" to "value";
     */
    
    uint32_t reg = *((uint32_t *) (&bios->data[offset + 1]));
    uint32_t value = *((uint32_t *) (&bios->data[offset + 5]));

	if (iexec->execute) {
		DEBUGPRINT(5, "0x%04X: REG: 0x%08X, VALUE: 0x%08X\n", 
				offset, reg, value);

		if (DEBUGLEVEL >= 6) {
			uint32_t tmpval;
			nv32_rd(reg, &tmpval);
			DEBUGPRINT(6, "0x%04X: CURRENT VALUE IS: 0x%08X\n", offset, tmpval);
		}

#ifdef PERFORM_WRITE
		nv32_wr(reg, value);
#endif
		
		/* Assign: reg = value */
	}
    return TRUE;
}
