#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "nvsoftboot.h"
#include "bios.h"
#include "parser.h"
#include "initcmds.h"
#include "nvclock.h"


static init_tbl_entry_t itbl_entry[] = {
   /* command name                   , id  , length  , command handler       */
    { "INIT_IO_RESTRICT_PROG"        , 0x32, 43      , init_io_restrict_prog },
    { "INIT_REPEAT"                  , 0x33, 2       , init_repeat           },
    { "INIT_END_REPEAT"              , 0x36, 1       , init_end_repeat       },
    { "INIT_COPY"                    , 0x37, 11      , init_copy             },
    { "INIT_NOT"                     , 0x38, 1       , init_not              },
    { "INIT_IO_FLAG_CONDITION"       , 0x39, 2       , init_io_flag_condition},
    { "INIT_IO_RESTRICT_PLL"         , 0x4A, 43      , init_io_restrict_pll  },
    { "INIT_PLL"                     , 0x4B, 9       , init_pll              },
    { "INIT_CR_INDEX_ADDRESS_LATCHED", 0x51, VARIABLE, init_cr_idx_adr_latch },
    { "INIT_CR"                      , 0x52, 4       , init_cr               },
    { "INIT_ZM_CR"                   , 0x53, 3       , init_zm_cr            },
    { "INIT_ZM_CR_GROUP"             , 0x54, VARIABLE, init_zm_cr_group      },
    { "INIT_CONDITION_TIME"          , 0x56, 3       , init_condition_time   },
    { "INIT_ZM_REG_SEQUENCE"         , 0x58, VARIABLE, init_zm_reg_sequence  },
    { "INIT_SUB_DIRECT"              , 0x5B, 3       , init_sub_direct       },
    { "INIT_COPY_NV_REG"             , 0x5F, 22      , init_copy_nv_reg      },
    { "INIT_ZM_INDEX_IO"             , 0x62, 5       , init_zm_index_io      },
    { "INIT_COMPUTE_MEM"             , 0x63, 1       , init_compute_mem      },
    { "INIT_RESET"                   , 0x65, 13      , init_reset            },
    { "INIT_SUB"                     , 0x6B, 2       , init_sub              },
    { "INIT_NV_REG"                  , 0x6E, 13      , init_nv_reg           },
    { "INIT_MACRO"                   , 0x6F, 2       , init_macro            },
    { "INIT_DONE"                    , 0x71, 1       , init_done             },
    { "INIT_RESUME"                  , 0x72, 1       , init_resume           },
    { "INIT_TIME"                    , 0x74, 3       , init_time             },
    { "INIT_CONDITION"               , 0x75, 2       , init_condition        },
    { "INIT_INDEX_IO"                , 0x78, 6       , init_index_io         },
    { "INIT_ZM_REG"                  , 0x7A, 9       , init_zm_reg           },
    { 0                              , 0   , 0       , 0                     }
};
#if 0
status_t find_macro_condition_offsets(bios_t *bios)
{
    /* Fills the bios_t struct with information about
     * macro and condition offsets. 
     */

    /* FIXME: Rewrite this function when the proper way to
     * locate macros and conditions is found.
     *
     * In the BIOS:es i have checked there is a "00/00/00"
     * almost right after the BIT structure ends. After the
     * string there is a 0x02 and a couple of 0x00:s. When the
     * 0x00:s end a segment with data starts. The macros 
     * location is found on this segment at an offset of 14 bytes 
     * from the segment start and the condition offset is
     * found at an offset of 16 bytes. Both these values 
     * are 16 bits.
     */

    uint16_t offset = findstr(bios, "00/00/00");

    if (offset == 0 || offset > (bios->length - 100)) {
        DEBUGPRINT(1, "Unable to locate 00/00/00 after BIT structure\n");
        return FAILURE;
    }
    /* move beyond the found the "00/00/00" and the trailing 0x02 */
    offset += 9;

    /* move past the 0x00:s */
    while ((offset < bios->length) && (bios->data[offset] == 0x0))
        offset++;

    if (offset > (bios->length - 18))
        return FAILURE;

    bios->macro_offset = *((uint16_t *) (&bios->data[offset + 14]));    
    bios->condition_offset = *((uint16_t *) (&bios->data[offset + 16]));    

    DEBUGPRINT(5, "Found macro offset: 0x%04X\n", bios->macro_offset);
    DEBUGPRINT(5, "Found condition offset: 0x%04X\n", bios->condition_offset);

    return SUCCESS; 
}
#endif

bios_t *load_bios(char *filename)
{

    int fd;
    bios_t *bios;
    struct stat statbuf;
    
    if (!(bios = (bios_t *) malloc(sizeof(bios_t)))) {
        DEBUGPRINT(1, "Unable to alloc memory for BIOS stucture.\n");
        return NULL;
    }
    
    if ((fd = open(filename, O_RDONLY)) == -1) {
        DEBUGPRINT(1, "Unable to open %s.\n", filename);
        return NULL;
    }
    
    if (fstat(fd, &statbuf) != 0) {
        DEBUGPRINT(1, "Unable to stat %s.\n", filename);
        return NULL;
    }
   
    bios->length = statbuf.st_size;
    
    if (!(bios->data = (unsigned char *) malloc(bios->length))) {
        DEBUGPRINT(1, "Unable to alloc memory for BIOS data.\n");
        return NULL;
    }
    
    if (read(fd, bios->data, bios->length) != bios->length) {
        DEBUGPRINT(1, "Unable to read BIOS.\n");
        return NULL;
    }
    
    close(fd);
    
    return bios;
}

void free_bios(bios_t *bios)
{
    free(bios->data);
    free(bios);
}

unsigned int get_init_table_entry_length(bios_t *bios, unsigned int offset, int i)
{
    /* Calculates the length of a given init table entry. */
    unsigned int length; 
    
    if (itbl_entry[i].length == VARIABLE) {
        switch (itbl_entry[i].id) {
            case 0x51:
                length = 5 + bios->data[offset + 4];
                break;
            case 0x54:
                length = 2 + bios->data[offset + 1] * 2;
                break;
            case 0x58:
                length = 6 + bios->data[offset + 5] * 4;
        }
    } else 
        length = itbl_entry[i].length;
    
    return length;
}

void parse_init_table(bios_t *bios, unsigned int offset, init_exec_t *iexec)
{
    
    /* Parses all commands in a init table. */
    
    /* We start out executing all commands found in the
     * init table. Some op codes may change the status
     * of this variable to SKIP, which will cause
     * the following op codes to perform no operation until
     * the value is changed back to EXECUTE.
     */
    unsigned char id;
    int i;
    
    
    
    /* Loop as long as INIT_DONE (command id 0x71) has not been found
     * (and offset < bios length just in case... ) */
    while (((id = bios->data[offset]) != 0x71) && offset < bios->length) {
        /* Find matching id in itbl_entry */
        for (i = 0; itbl_entry[i].name && (itbl_entry[i].id != id); i++)
            ;
        
        if (itbl_entry[i].name) {
            
            DEBUGPRINT(3, "0x%04X:  [ (0x%02X) -  %s ]\n", offset, 
                    itbl_entry[i].id, itbl_entry[i].name);
            
            /* execute eventual command handler */
            if (itbl_entry[i].handler)
                if (!(*itbl_entry[i].handler)(bios, offset, iexec))
                    break;
        
        } else {
            DEBUGPRINT(2, "0x%04X: Init table command not found: 0x%02X\n", 
                    offset, id);
        }
        
        /* Add the offset of the current command including all data
         * of that command. The offset will then be pointing on the
         * next op code.
         */
        offset += get_init_table_entry_length(bios, offset, i);
    }
}

void parse_init_tables(bios_t *bios) {
    
    /* Loops and calls parse_init_table() for each present table. */
    
    int i = 0;
    uint16_t table;
    init_exec_t iexec = {TRUE, FALSE};
    
    while (table = *((uint16_t *) (&bios->data[bios->init_tbls_offset + i]))) {
        
        DEBUGPRINT(2, "0x%04X: Parsing init table %d\n", 
                table, i / 2);
        
        DEBUGPRINT(3, "0x%04X: ------ EXECUTING FOLLOWING COMMANDS ------\n");
        
        parse_init_table(bios, table, &iexec);         
        i += 2;
    }
}
status_t parse_bit_init_tbl_entry(bios_t *bios, bit_entry_t *bitentry) {
    
    /* Parses the init table segment that the bit entry points to.
     * Starting at bitentry->offset: 
     * 
     * offset + 0  (16 bits): offset of init tables
     * offset + 2  (16 bits): unknown
     * offset + 4  (16 bits): macro offset
     * offset + 6  (16 bits): condition offset
     * offset + 8  (16 bits): io flag condition offset (?)
     * offset + 10 (16 bits): io flag condition offset (?)
     *
     * offset + 8 and offset + 10 seems to contain the same
     * offsets on all bioses i have checked. Don't know which
     * one is the correct, therefore this code will bail out
     * if the two values are not the same.
     */
    
    if (bitentry->length < 12) {
        DEBUGPRINT(1, "Unable to regocnize BIT init table entry.\n");
        return FAILURE;
    }
    
    bios->init_tbls_offset = *((uint16_t *) (&bios->data[bitentry->offset]));
    /* XXX: ram config at offset + 2 ?? */
    bios->ram_table_offset = *((uint16_t *) (&bios->data[bitentry->offset + 2]));
    bios->macro_offset = *((uint16_t *) (&bios->data[bitentry->offset + 4]));
    bios->condition_offset = 
        *((uint16_t *) (&bios->data[bitentry->offset + 6]));
    
    if (*((uint16_t *) (&bios->data[bitentry->offset + 8])) != 
            *((uint16_t *) (&bios->data[bitentry->offset + 10]))) {
        DEBUGPRINT(1, "Unable to find IO flag condition offset.\n");
        return FAILURE;
    }
    
    bios->io_flag_condition_offset =
        *((uint16_t *) (&bios->data[bitentry->offset + 8]));
    
    parse_init_tables(bios);
    
    return SUCCESS;
}

void parse_bit_structure(bios_t *bios, unsigned int offset) {
    
    bit_entry_t *bitentry;
    char done = 0;
    
    while (!done) {
        bitentry = (bit_entry_t *) &bios->data[offset];
        
        DEBUGPRINT(2, "0x%04X: Found BIT command with id 0x%02X\n", 
                offset, bitentry->id[0]); 
        
        switch (bitentry->id[0]) {
            case 0:
                /* id[0] = 0 and id[1] = 0  ==> end of BIT struture */
                if (bitentry->id[1] == 0)
                    done = 1;
                break;
            case 'I':
                DEBUGPRINT(2, 
                        "0x%04X: Found init table entry in BIT structure.\n", 
                        offset);

                parse_bit_init_tbl_entry(bios, bitentry);
                //parse_init_tables(bios, bitentry);
                break;
        }

        offset += sizeof(bit_entry_t);
    }
}


status_t parse_bios(bios_t *bios) {
    
    unsigned int bit_offset;
    
    /* check for BIOS signature */
    if (!(bios->data[0] == 0x55 && bios->data[1] == 0xAA)) {
        DEBUGPRINT(1, "BIOS signature not found!\n");
        return FAILURE;
    }
    
    DEBUGPRINT(2, "BIOS signature found.\n");
    
    /* check for BIT signature */
    if (!(bit_offset = findstr(bios, "BIT"))) {
        DEBUGPRINT(1, "BIT signature not found!\n");
        return FAILURE;
    }
    
    DEBUGPRINT(2, "BIT signature found.\n");
    
    parse_bit_structure(bios, bit_offset + 4);
    
    return SUCCESS;
}


unsigned int findstr(bios_t *bios, unsigned char *str) {
    
    int i;
    int len = strlen(str);
    
    for (i = 2; i < bios->length; i++)
        if (strncmp(&bios->data[i], str, len) == 0)
            return i;

    return 0;
}

int main(int argc, char *argv[]) {
    
    int i;
    bios_t *bios;
   
    if (!init_nvclock()) {
        fprintf(stderr, "init_nvclock() failed!\n");
        exit(1);
    }

    sleep(1);

    if (!set_card(0)) {
        fprintf(stderr, "set_card() failed!\n");
        exit(2);
    }
 
    sleep(1);

    NVLockUnlock(0);
    printf("0x%08X\n", nv_card.PMC[0x15f0 / 4]);
	sleep(1);
	nv_card.PMC[0x15f0 / 4] = 0x00150015;
	sleep(1);
	printf("0x%08X\n", nv_card.PMC[0x15f0 / 4]);
    if (!(bios = load_bios(argv[1])))
        return 1;
    
    parse_bios(bios);
    
    free_bios(bios);
        
    NVLockUnlock(1);
    return 0;
}

