#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <fcntl.h>

#include "bios.h"
#include "parser.h"
#include "initcmds.h"

static init_tbl_entry_t itbl_entry[] = {
   /* command name                   , id  , length  , command handler       */
    { "INIT_IO_RESTRICT_PROG"        , 0x32, 43      , init_io_restrict_prog },
    { "INIT_REPEAT"                  , 0x33, 2       , init_repeat           },
    { "INIT_END_REPEAT"              , 0x36, 1       , init_end_repeat       },
    { "INIT_COPY"                    , 0x37, 11      , init_copy             },
    { "INIT_NOT"                     , 0x38, 1       , init_not              },
    { "UNKNOWN"                      , 0x39, 2       , NULL                  },
    { "INIT_IO_RESTRICT_PLL"         , 0x4A, 43      , init_io_restrict_pll  },
    { "INIT_PLL"                     , 0x4B, 9       , init_pll              },
    { "INIT_CR_INDEX_ADDRESS_LATCHED", 0x51, VARIABLE, init_cr_idx_adr_latch },
    { "INIT_CR"                      , 0x52, 4       , init_cr               },
    { "INIT_ZM_CR"                   , 0x53, 3       , init_zm_cr            },
    { "INIT_ZM_CR_GROUP"             , 0x54, VARIABLE, init_zm_cr_group      },
    { "INIT_CONDITION_TIME"          , 0x56, 3       , init_condition_time   },
    { "INIT_ZM_REG_SEQUENCE"         , 0x58, VARIABLE, init_zm_reg_sequence  },
    { "INIT_SUB_DIRECT"              , 0x5B, 3       , init_sub_direct       },
    { "INIT_COPY_NV_REG"             , 0x5F, 22      , init_copy_nv_reg      },
    { "INIT_ZM_INDEX_IO"             , 0x62, 5       , init_zm_index_io      },
    { "INIT_COMPUTE_MEM"             , 0x63, 1       , init_compute_mem      },
    { "INIT_RESET"                   , 0x65, 13      , init_reset            },
    { "INIT_SUB"                     , 0x6B, 2       , init_sub              },
    { "INIT_NV_REG"                  , 0x6E, 13      , init_nv_reg           },
    { "INIT_MACRO"                   , 0x6F, 2       , init_macro            },
    { "INIT_DONE"                    , 0x71, 1       , init_done             },
    { "INIT_RESUME"                  , 0x72, 1       , init_resume           },
    { "INIT_TIME"                    , 0x74, 3       , init_time             },
    { "INIT_CONDITION"               , 0x75, 2       , init_condition        },
    { "INIT_INDEX_IO"                , 0x78, 6       , init_index_io         },
    { "INIT_ZM_REG"                  , 0x7A, 9       , init_zm_reg           },
    { 0                              , 0   , 0       , 0                     }
};

bios_t *load_bios(char *filename) {

    int fd;
    bios_t *bios;
    struct stat statbuf;
    
    if (!(bios = (bios_t *) malloc(sizeof(bios_t)))) {
        DEBUGPRINT(1, "Unable to alloc memory for BIOS stucture.\n");
        return NULL;
    }
    
    if ((fd = open(filename, O_RDONLY)) == -1) {
        DEBUGPRINT(1, "Unable to open %s.\n", filename);
        return NULL;
    }
    
    if (fstat(fd, &statbuf) != 0) {
        DEBUGPRINT(1, "Unable to stat %s.\n", filename);
        return NULL;
    }
   
    bios->length = statbuf.st_size;
    
    if (!(bios->data = (unsigned char *) malloc(bios->length))) {
        DEBUGPRINT(1, "Unable to alloc memory for BIOS data.\n");
        return NULL;
    }
    
    if (read(fd, bios->data, bios->length) != bios->length) {
        DEBUGPRINT(1, "Unable to read BIOS.\n");
        return NULL;
    }
    
    close(fd);
    
    return bios;
}

void free_bios(bios_t *bios) {
    free(bios->data);
    free(bios);
}

unsigned int get_init_table_entry_length(bios_t *bios, unsigned int offset, int i) {
    
    /* Calculates the length of a given init table entry. */
    unsigned int length; 
    
    if (itbl_entry[i].length == VARIABLE) {
        switch (itbl_entry[i].id) {
            case 0x51:
                length = 5 + bios->data[offset + 4];
                break;
            case 0x54:
                length = 2 + bios->data[offset + 1] * 2;
                break;
            case 0x58:
                length = 6 + bios->data[offset + 5] * 4;
        }
    } else {
        length = itbl_entry[i].length;
    }

    return length;
}
void parse_init_table(bios_t *bios, unsigned int offset) {
    
    /* Parses all commands in a init table. */
    unsigned char id;
    int i;
    
    /* Loop as long as INIT_DONE (command id 0x71) has not been found
     * (and offset < bios length just in case... ) */
    while (((id = bios->data[offset]) != 0x71) && offset < bios->length) {
        /* Find matching id in itbl_entry */
        for (i = 0; itbl_entry[i].name && (itbl_entry[i].id != id); i++)
            ;
        
        if (itbl_entry[i].name) {
            DEBUGPRINT(3, "Init table command 0x%02X: %s\n", itbl_entry[i].id,
                    itbl_entry[i].name);
             itbl_entry[i].used = 1;        
            
             /* execute command eventual command handler */
             if (itbl_entry[i].handler)
                 (*itbl_entry[i].handler)(bios, offset + 1);
        
        } else {
            DEBUGPRINT(2, "Init table command not found: 0x%02X\n", id);
        }
        
        offset += get_init_table_entry_length(bios, offset, i);
    }
}

void parse_init_tables(bios_t *bios, bit_entry_t *bitentry) {
    
    /* Loops and calls parse_init_table() for each present table. */
    
    int i;
    uint16_t table;
    uint16_t offset = *((uint16_t *) (&bios->data[bitentry->offset])); 
    
    DEBUGPRINT(2, "Number of tables: %d\n", bitentry->length / 2);
    
    for (i = 0; i < (bitentry->length / 2); i++) {
        
        DEBUGPRINT(2, "Parsing init table %d.\n", i + 1);
        table = *((uint16_t *) (&bios->data[offset]));
        
        if (table) {
            DEBUGPRINT(2, "Table %d contains data.\n", i + 1);
            parse_init_table(bios, table);         
        }
        
        offset += 2;
    }
}

int parse_bit_structure(bios_t *bios, unsigned int offset) {
    
    bit_entry_t *bitentry;
    char done = 0;
    
    while (!done) {
        bitentry = (bit_entry_t *) &bios->data[offset];
        DEBUGPRINT(2, "Found BIT command with id 0x%02X\n", bitentry->id[0]); 
        switch (bitentry->id[0]) {
            case 0:
                /* id[0] = 0 and id[1] = 0  ==> end of BIT struture */
                if (bitentry->id[1] == 0)
                    done = 1;
                break;
            case 'I':
                DEBUGPRINT(2, "Found init table entry in BIT structure.\n");
                parse_init_tables(bios, bitentry);
                break;
        }

        offset += sizeof(bit_entry_t);
    }

    return 1;
}


int parse_bios(bios_t *bios) {
    
    unsigned int bit_offset;
    
    /* check for BIOS signature */
    if (!(bios->data[0] == 0x55 && bios->data[1] == 0xAA)) {
        DEBUGPRINT(1, "BIOS signature not found!\n");
        return 0;
    }
    
    DEBUGPRINT(2, "BIOS signature found.\n");
    
    /* check for BIT signature */
    if (!(bit_offset = findstr(bios, "BIT"))) {
        DEBUGPRINT(1, "BIT signature not found!\n");
        return 0;
    }
    
    DEBUGPRINT(2, "BIT signature found.\n");
    
    if (!(parse_bit_structure(bios, bit_offset + 4))) {
        DEBUGPRINT(1, "BIT structure parsing failed.\n");
        return 0;
    }
    
    return 1;
}


unsigned int findstr(bios_t *bios, unsigned char *str) {
    
    int i;
    int len = strlen(str);
    
    for (i = 2; i < bios->length; i++)
        if (strncmp(&bios->data[i], str, len) == 0)
            return i;

    return 0;
}

int main(int argc, char *argv[]) {
    
    int i;
    int ones = 0;
    int zeros = 0;
    bios_t *bios;
    
    if (!(bios = load_bios(argv[1])))
        return 1;
    
    parse_bios(bios);
    
    for (i = 0; itbl_entry[i].name; i++) {
        if (itbl_entry[i].used) {
            ones++;
        } else {
            printf("%s NOT used.\n", itbl_entry[i].name);
            zeros++;
        }
    }

    printf("%d commands used\n%d commands NOT used\n", ones, zeros);
    free_bios(bios);
        
    return 0;
}

